import numpy as np
import PIL.Image


def rgb2gray(rgb):
    # type: (np.ndarray) -> np.ndarray
    '''Covnert rgb to gray.

    Parameters
    ----------
    rgb: numpy.ndarray, (H, W, 3), np.uint8
        Input rgb image.

    Returns
    -------
    gray: numpy.ndarray, (H, W)
        Output gray image.
    '''
    assert rgb.ndim == 3, 'rgb must be 3 dimensional'
    assert rgb.shape[2] == 3, 'rgb shape must be (H, W, 3)'
    assert rgb.dtype == np.uint8, 'rgb dtype must be np.uint8'

    gray = PIL.Image.fromarray(rgb)
    gray = gray.convert('L')
    gray = np.asarray(gray)
    return gray


def gray2rgb(gray):
    # type: (np.ndarray) -> np.ndarray
    '''Covnert gray to rgb.

    Parameters
    ----------
    gray: numpy.ndarray, (H, W), np.uint8
        Input gray image.

    Returns
    -------
    rgb: numpy.ndarray, (H, W, 3), np.uint8
        Output rgb image.
    '''
    assert gray.ndim == 2, 'gray must be 2 dimensional'
    assert gray.dtype == np.uint8, 'gray dtype must be np.uint8'

    rgb = gray[:, :, None].repeat(3, axis=2)
    return rgb


def rgb2rgba(rgb):
    # type: (np.ndarray) -> np.ndarray
    '''Convert rgb to rgba.

    Parameters
    ----------
    rgb: numpy.ndarray, (H, W, 3), np.uint8
        Input rgb image.

    Returns
    -------
    rgba: numpy.ndarray, (H, W, 4), np.uint8
        Output rgba image.
    '''
    assert rgb.ndim == 3, 'rgb must be 3 dimensional'
    assert rgb.shape[2] == 3, 'rgb shape must be (H, W, 3)'
    assert rgb.dtype == np.uint8, 'rgb dtype must be np.uint8'

    a = np.full(rgb.shape[:2], 255, dtype=np.uint8)
    rgba = np.dstack((rgb, a))
    return rgba



# flake8: noqa

import os.path as osp

import numpy as np


here = osp.dirname(osp.abspath(__file__))


def arc2017():
    data_file = osp.join(here, '1532900700692405455.npz')
    data = np.load(data_file)
    data = dict(data)

    # compose masks to class label image
    class_label = np.full(data['rgb'].shape[:2], 0, dtype=np.int32)
    for l, mask in zip(data['labels'], data['masks']):
        class_label[mask == 1] = l
    data['class_label'] = class_label

    names_file = osp.join(here, 'class_names.txt')
    with open(names_file) as f:
        class_names = [name.strip() for name in f]
    data['class_names'] = class_names

    data['res4'] = np.load(osp.join(here, 'res4.npz'))['res4']

    return data



import numpy as np

from . import color as color_module
from . import draw as draw_module
from . import label as label_module


def mask_to_bbox(masks):
    bboxes = np.zeros((len(masks), 4), dtype=float)
    for i, mask in enumerate(masks):
        where = np.argwhere(mask)
        (y1, x1), (y2, x2) = where.min(0), where.max(0) + 1
        bbox = y1, x1, y2, x2
        bboxes[i] = bbox
    return bboxes


def instances2rgb(
    image,
    labels,
    bboxes=None,
    masks=None,
    captions=None,
    font_size=25,
    line_width=5,
    alpha=0.7,
):
    assert isinstance(image, np.ndarray)
    assert image.dtype == np.uint8
    assert image.ndim == 3

    assert all(l >= 0 for l in labels)

    n_instance = len(labels)

    if masks is None:
        assert bboxes is not None
        masks = [None] * n_instance
    if bboxes is None:
        assert masks is not None
        bboxes = mask_to_bbox(masks)
    if captions is None:
        captions = [None] * n_instance

    assert len(masks) == len(bboxes) == len(captions) == n_instance

    colormap = label_module.label_colormap()
    colormap = (colormap * 255).round().astype(np.uint8)

    dst = image
    image_gray = color_module.gray2rgb(color_module.rgb2gray(image))

    for instance_id in range(n_instance):
        mask = masks[instance_id]

        if mask is None:
            continue

        color_ins = colormap[1:][instance_id % len(colormap[1:])]

        maskviz = mask[:, :, None] * color_ins.astype(float)
        dst = dst.copy()
        dst[mask] = (
            (1 - alpha) * image_gray[mask].astype(float) +
            alpha * maskviz[mask]
        )

        try:
            import skimage.segmentation
            boundary = skimage.segmentation.find_boundaries(
                mask, connectivity=2
            )
            dst[boundary] = (200, 200, 200)
        except ImportError:
            pass

    for instance_id in range(n_instance):
        bbox = bboxes[instance_id]
        label = labels[instance_id]
        caption = captions[instance_id]

        color_cls = colormap[label % len(colormap)]

        y1, x1, y2, x2 = bbox
        aabb1 = np.array([y1, x1], dtype=int)
        aabb2 = np.array([y2, x2], dtype=int)
        dst = draw_module.rectangle(
            dst,
            aabb1,
            aabb2,
            color=tuple(color_cls),
            width=line_width,
        )

        if caption is not None:
            height, width = draw_module.text_size(text=caption, size=font_size)
            dst = draw_module.rectangle(
                dst,
                aabb1,
                aabb1 + [height, width],
                color=tuple(color_cls),
                fill=tuple(color_cls),
            )
            dst = draw_module.text(
                dst,
                (y1, x1),
                text=caption,
                color=(255, 255, 255),
                size=font_size,
            )
    return dst



import numpy as np  # NOQA
import PIL.Image

try:
    import pyglet
except ImportError:
    pyglet = None


def check_pyglet_available():
    if pyglet is None:
        raise ImportError(
            'pyglet is not installed, run following: pip install pyglet'
        )


def pyglet_imshow(image, caption=None):
    # type: (np.ndarray) -> None
    check_pyglet_available()

    image = PIL.Image.fromarray(image)
    image = pyglet.image.ImageData(
        width=image.width,
        height=image.height,
        format=image.mode,
        data=image.tobytes(),
        pitch=- image.width * len(image.mode),
    )

    window = pyglet.window.Window(
        width=image.width,
        height=image.height,
        caption=caption,
        # resizable=True,
    )
    sprite = pyglet.sprite.Sprite(image)

    @window.event
    def on_draw():
        window.clear()
        sprite.draw()

    @window.event()
    def on_key_press(symbol, modifiers):
        if modifiers == 0:
            if symbol == pyglet.window.key.Q:
                window.close()

    # @window.event()
    # def on_resize(width, height):
    #     scale_w = 1. * width / image.width
    #     scale_h = 1. * height / image.height
    #
    #     image_scale = min(scale_w, scale_h)
    #     image_width = int(round(image.width * sprite.scale))
    #     image_height = int(round(image.height * sprite.scale))
    #
    #     sprite.update(
    #         x=(width - image_width) / 2.0,
    #         y=(height - image_height) / 2.0,
    #         scale=image_scale,
    #     )


def pyglet_run():
    check_pyglet_available()

    return pyglet.app.run()



import numpy as np


def pyplot_fig2arr(fig):
    fig.canvas.draw()
    arr = np.frombuffer(fig.canvas.tostring_rgb(), dtype=np.uint8)
    width, height = fig.canvas.get_width_height()
    arr = arr.reshape((height, width, 3))
    return arr



import os.path as osp

import numpy as np


here = osp.dirname(osp.abspath(__file__))


def read_pose_file(filename):
    with open(filename, 'r') as f:
        transforms = []
        for one_line in f:
            one_line = one_line.split(' ')
            Rt = [float(pose) for pose in one_line] + [0, 0, 0, 1]

            Rt = np.reshape(np.array(Rt), (4, 4))
            assert abs(Rt[3].sum() - 1) < 1e-5
            transforms.append(Rt)
    return transforms


def kitti_odometry():
    # http://www.cvlibs.net/datasets/kitti/eval_odometry.php
    pose_file = osp.join(here, 'odometry/00.txt')
    transforms = read_pose_file(pose_file)
    data = {'transforms': transforms}
    return data



import matplotlib.cm
import numpy as np

from .normalize import normalize


class Depth2RGB(object):

    """Convert depth array to rgb."""

    def __init__(self, min_value=None, max_value=None, colormap='jet'):
        self._min_value = min_value
        self._max_value = max_value

        assert hasattr(matplotlib.cm, colormap), \
            'unsupported colormap: {}'.format(colormap)
        self._colormap = colormap

    @property
    def min_value(self):
        return self._min_value

    @property
    def max_value(self):
        return self._max_value

    def __call__(self, depth, dtype=np.uint8):
        assert depth.ndim == 2, 'depth image must be 2 dimensional'
        assert np.issubdtype(depth.dtype, np.floating), \
            'depth dtype must be float'

        normalized = normalize(
            depth, min_value=self._min_value, max_value=self._max_value
        )

        isnan = np.isnan(normalized)
        normalized[isnan] = 0

        colormap_func = getattr(matplotlib.cm, self._colormap)
        rgb = colormap_func(normalized)[:, :, :3]
        rgb[isnan] = (0, 0, 0)

        if dtype == np.uint8:
            rgb = (rgb * 255).round().astype(np.uint8)
        else:
            assert np.issubdtype(dtype, np.floating)
            rgb = rgb.astype(dtype)

        return rgb


def depth2rgb(
    depth,
    dtype=np.uint8,
    min_value=None,
    max_value=None,
    colormap='jet',
):
    '''Convert depth to rgb.

    Parameters
    ----------
    depth: numpy.ndarray, (H, W), float
        Depth image.
    dtype: numpy.dtype
        Dtype of output image. default: np.uint8
    min_value: float, optional
        Minimum value for colorizing.
    max_value: float, optional
        Maximum value for colorizing.
    colormap: str, optional
        Colormap, default: 'jet'.

    Returns
    -------
    rgb: numpy.ndarray, (H, W, 3), np.uint8
        Output colorized image.
    '''
    return Depth2RGB(min_value, max_value, colormap)(depth, dtype)



try:
    import cv2
except ImportError:
    cv2 = None


def check_cv2_available():
    if cv2 is None:
        raise ImportError(
            'cv2 is not installed, run following: pip install opencv-python'
        )


def cv_imshow(image, window_name=''):
    check_cv2_available()

    return cv2.imshow(window_name, image[:, :, ::-1])


def cv_waitkey(msec=0):
    check_cv2_available()

    return cv2.waitKey(msec)



import numpy as np
import PIL.Image


def imread(filename):
    # type: (str) -> np.ndarray
    return np.asarray(PIL.Image.open(filename))


def imsave(filename, arr):
    # type: (str, np.ndarray) -> None
    return PIL.Image.fromarray(arr).save(filename)



# flake8: noqa

from .base import imread
from .base import imsave

from .opencv import cv_imshow
from .opencv import cv_waitkey

from .pyplot import pyplot_fig2arr

from ._pyglet import pyglet_imshow
from ._pyglet import pyglet_run



import numpy as np

from .resize import resize


def centerize(src, shape, cval=None, return_mask=False):
    """Centerize image for specified image size

    Parameters
    ----------
    src: numpy.ndarray
        Image to centerize
    shape: tuple of int
        Image shape (height, width) or (height, width, channel)
    cval: numpy.ndarray
        Color to be filled in the blank.
    return_mask: numpy.ndarray
        Mask for centerized image.

    Returns
    -------
    dst: numpy.ndarray
        Centerized image.
    """
    if src.shape[:2] == shape[:2]:
        if return_mask:
            return src, np.ones(shape[:2], dtype=bool)
        else:
            return src

    if len(shape) != src.ndim:
        shape = list(shape) + [src.shape[2]]

    dst = np.zeros(shape, dtype=src.dtype)
    if cval:
        dst[:, :] = cval

    src_h, src_w = src.shape[:2]
    scale_h, scale_w = 1. * shape[0] / src_h, 1. * shape[1] / src_w
    scale = min(scale_h, scale_w)
    dst_h, dst_w = int(round(src_h * scale)), int(round(src_w * scale))
    src = resize(src, height=dst_h, width=dst_w)

    ph, pw = 0, 0
    h, w = src.shape[:2]
    dst_h, dst_w = shape[:2]
    if h < dst_h:
        ph = (dst_h - h) // 2
    if w < dst_w:
        pw = (dst_w - w) // 2
    dst[ph:ph + h, pw:pw + w] = src

    if return_mask:
        mask = np.zeros(shape[:2], dtype=bool)
        mask[ph:ph + h, pw:pw + w] = True
        return dst, mask
    else:
        return dst



try:
    import cv2
except ImportError:
    cv2 = None

import numpy as np
import PIL.Image


def _resize_pillow(src, height, width, interpolation):
    if interpolation == 'linear':
        interpolation = PIL.Image.LINEAR
    elif interpolation == 'nearest':
        interpolation = PIL.Image.NEAREST
    else:
        raise ValueError(
            'unsupported interpolation: {}'.format(interpolation)
        )

    if src.dtype == np.uint8:
        dst = PIL.Image.fromarray(src)
        dst = dst.resize((width, height), resample=interpolation)
        dst = np.asarray(dst)
    else:
        assert np.issubdtype(src.dtype, np.floating)
        ndim = src.ndim
        if ndim == 2:
            src = src[:, :, None]

        C = src.shape[2]
        dst = np.zeros((height, width, C), dtype=src.dtype)
        for c in range(C):
            src_c = src[:, :, c]
            src_c = PIL.Image.fromarray(src_c)
            dst[:, :, c] = src_c.resize(
                (width, height), resample=interpolation
            )

        if ndim == 2:
            dst = dst[:, :, 0]
    return dst


def _resize_opencv(src, height, width, interpolation):
    if interpolation == 'linear':
        interpolation = cv2.INTER_LINEAR
    elif interpolation == 'nearest':
        interpolation = cv2.INTER_NEAREST
    else:
        raise ValueError(
            'unsupported interpolation: {}'.format(interpolation)
        )

    dst = cv2.resize(src, (width, height), interpolation=interpolation)
    return dst


def resize(
    src,
    height=None,
    width=None,
    interpolation='linear',
    backend='auto',
):
    if not isinstance(src, np.ndarray):
        raise TypeError('src type must be numpy.ndarray')

    if backend == 'auto':
        backend = 'pillow' if cv2 is None else 'opencv'

    src_height, src_width = src.shape[:2]
    if isinstance(width, float):
        scale_width = width
        width = int(round(scale_width * src_width))
    if isinstance(height, float):
        scale_height = height
        height = int(round(scale_height * src_height))
    if height is None:
        assert width is not None
        scale_height = 1. * width / src_width
        height = int(round(scale_height * src_height))
    if width is None:
        assert height is not None
        scale_width = 1. * height / src_height
        width = int(round(scale_width * src_width))

    if backend == 'pillow':
        dst = _resize_pillow(src, height, width, interpolation)
    elif backend == 'opencv':
        dst = _resize_opencv(src, height, width, interpolation)
    else:
        raise ValueError('unsupported backend: {}'.format(backend))

    return dst



from . import normalize

import numpy as np


class Nchannel2RGB(object):

    """Convert nchannel array to rgb by PCA."""

    def __init__(self, min_value=None, max_value=None, pca=None):
        self._min_value = min_value
        self._max_value = max_value
        self._pca = pca

    @property
    def pca(self):
        return self._pca

    @property
    def min_value(self):
        return self._min_value

    @property
    def max_value(self):
        return self._max_value

    def __call__(self, nchannel, dtype=np.uint8):
        import sklearn.decomposition

        assert nchannel.ndim == 3, 'nchannel.ndim must be 3'
        assert np.issubdtype(nchannel.dtype, np.floating), \
            'nchannel.dtype must be floating'
        H, W, D = nchannel.shape

        if self._min_value is None:
            self._min_value = np.nanmin(nchannel, axis=(0, 1))
        if self._max_value is None:
            self._max_value = np.nanmax(nchannel, axis=(0, 1))
        dst = normalize.normalize(nchannel, self._min_value, self._max_value)

        dst = dst.reshape(-1, D)
        if self._pca is None:
            self._pca = sklearn.decomposition.PCA(n_components=3)
            dst = self._pca.fit_transform(dst)
        else:
            dst = self._pca.transform(dst)
        dst = dst.reshape(H, W, 3)

        if dtype == np.uint8:
            dst = (dst * 255).round().astype(np.uint8)
        else:
            assert np.issubdtype(dtype, np.floating)
            dst = dst.astype(dtype)

        return dst


def nchannel2rgb(
    nchannel, dtype=None, pca=None, min_value=None, max_value=None,
):
    return Nchannel2RGB(min_value, max_value, pca)(nchannel, dtype)



import warnings

import numpy as np


def normalize(src, min_value=None, max_value=None):
    if src.ndim == 2:
        D = 1
    else:
        assert src.ndim == 3, 'src ndim must be 2 or 3'
        D = src.shape[2]

    if min_value is None:
        min_value = np.nanmin(src, axis=(0, 1))
    min_value = np.atleast_1d(min_value).astype(float)
    assert min_value.shape == (D,)

    if max_value is None:
        max_value = np.nanmax(src, axis=(0, 1))
    max_value = np.atleast_1d(max_value).astype(float)
    assert max_value.shape == (D,)

    if np.isinf(min_value).any() or np.isinf(max_value).any():
        warnings.warn('some of min or max values are inf.')

    eps = np.finfo(src.dtype).eps
    issame = max_value == min_value
    min_value[issame] -= eps
    max_value[issame] += eps

    dst = np.zeros(src.shape, dtype=float)

    if src.ndim == 2:
        isnan = np.isnan(src)
    else:
        isnan = np.isnan(src).any(axis=2)
    dst[~isnan] = (
        1. * (src[~isnan] - min_value) / (max_value - min_value)
    )
    dst[isnan] = np.nan

    return dst



import os.path as osp

import matplotlib
import numpy as np
import PIL.Image
import PIL.ImageDraw
import PIL.ImageFont


def rectangle(src, aabb1, aabb2, color, fill=None, width=0):
    '''Draw rectangle on numpy array with Pillow.

    Parameters
    ----------
    src: numpy.ndarray
        Input image.
    aabb1, aabb2: (2,) array-like
        aabb1 is (y_min, x_min) and aabb2 is (y_max, x_max).
    color: (3,) array-like
        RGB color in uint8.
    fill: (3,) array-like, optional
        RGB color to fill the rectangle. None for no fill. (default: None)
    width: int, optional
        Rectangle line width. (default: 0)

    Returns
    -------
    dst: numpy.ndarray
        Output image.
    '''
    color = tuple(color)

    src_pil = PIL.Image.fromarray(src)
    draw = PIL.ImageDraw.ImageDraw(src_pil)

    y1, x1 = aabb1
    y2, x2 = aabb2
    draw.rectangle(xy=(x1, y1, x2, y2), fill=fill, outline=color, width=width)

    dst = np.asarray(src_pil)
    return dst


def _get_font(size):
    fonts_path = osp.join(
        osp.dirname(matplotlib.__file__), 'mpl-data/fonts/ttf'
    )
    font_path = osp.join(fonts_path, 'DejaVuSans.ttf')
    font = PIL.ImageFont.truetype(font=font_path, size=size)
    return font


def text_size(text, size):
    '''Get text size (height and width).

    Parameters
    ----------
    text: str
        Text.
    size: int
        Pixel font size.

    Returns
    -------
    height: int
        Text height.
    width: int
        Text width.
    '''
    font = _get_font(size)
    width, height = font.getsize(text)
    return height, width


def text(src, yx, text, color, size):
    '''Draw text on numpy array with Pillow.

    Parameters
    ----------
    src: numpy.ndarray
        Input image.
    yx: (2,) array-like
        Left top point of the text.
    text: str
        Text to draw.
    color: (3,) array-like
        Text RGB color in uint8
    size: int
        Text size in pixel.

    Returns
    -------
    dst: numpy.ndarray
        Output image.
    '''
    src_pil = PIL.Image.fromarray(src)
    draw = PIL.ImageDraw.ImageDraw(src_pil)

    y1, x1 = yx
    color = tuple(color)
    font = _get_font(size=size)
    draw.text(xy=(x1, y1), text=text, fill=color, font=font)

    dst = np.asarray(src_pil)
    return dst



# flake8: noqa

from . import data

from . import draw

from . import _io as io

# -----------------------------------------------------------------------------

# visualization targets
from .color import gray2rgb
from .color import rgb2gray
from .depth import Depth2RGB
from .depth import depth2rgb
from .flow import flow2rgb
from .instances import instances2rgb
from .label import label_colormap
from .label import label2rgb
from .nchannel import Nchannel2RGB
from .nchannel import nchannel2rgb
from .trajectory import plot_trajectory

# visualization operations
from .centerize import centerize
from .normalize import normalize
from .resize import resize
from .tile import tile
# from .overlay import overlay



import numpy as np

from . import color as color_module
from . import draw as draw_module


def label_colormap(n_label=256):
    # type: (int) -> np.ndarray
    def bitget(byteval, idx):
        return ((byteval & (1 << idx)) != 0)

    cmap = np.zeros((n_label, 3))
    for i in range(0, n_label):
        id = i
        r, g, b = 0, 0, 0
        for j in range(0, 8):
            r = np.bitwise_or(r, (bitget(id, 0) << 7 - j))
            g = np.bitwise_or(g, (bitget(id, 1) << 7 - j))
            b = np.bitwise_or(b, (bitget(id, 2) << 7 - j))
            id = (id >> 3)
        cmap[i, 0] = r
        cmap[i, 1] = g
        cmap[i, 2] = b
    cmap = cmap.astype(np.float32) / 255
    return cmap


def label2rgb(
    label,
    label_names=None,
    n_labels=None,
    font_size=30,
    thresh_suppress=0,
):
    if n_labels is None:
        if label_names:
            n_labels = len(label_names)
        else:
            n_labels = np.max(label) + 1  # +1 for background label 0
    elif label_names:
        assert n_labels == len(label_names)
        assert np.max(label) < n_labels

    colormap = label_colormap(n_labels)
    colormap = (colormap * 255).astype(np.uint8)

    res = colormap[label]

    np.random.seed(1234)

    mask_unlabeled = label < 0
    res[mask_unlabeled] = \
        np.random.random(size=(mask_unlabeled.sum(), 3)) * 255

    if label_names is None:
        return res

    def get_fg_color(color):
        intensity = color_module.rgb2gray(color.reshape(1, 1, 3)).sum()
        if intensity > 170:
            return (0, 0, 0)
        return (255, 255, 255)

    def center_of_mass(mask):
        assert mask.ndim == 2 and mask.dtype == bool
        mask = 1. * mask / mask.sum()
        dx = np.sum(mask, 0)
        dy = np.sum(mask, 1)
        cx = np.sum(dx * np.arange(mask.shape[1]))
        cy = np.sum(dy * np.arange(mask.shape[0]))
        return cy, cx

    for l in np.unique(label):
        if l == -1:
            continue  # unlabeled

        mask = label == l
        if 1. * mask.sum() / mask.size < thresh_suppress:
            continue
        y, x = np.array(center_of_mass(mask), dtype=int)

        if label[y, x] != l:
            Y, X = np.where(mask)
            point_index = np.random.randint(0, len(Y))
            y, x = Y[point_index], X[point_index]

        text = label_names[l]
        height, width = draw_module.text_size(text, size=font_size)
        color = get_fg_color(res[y, x])
        res = draw_module.text(
            res,
            yx=(y - height // 2, x - width // 2),
            text=text,
            color=color,
            size=font_size,
        )

    return res
